import hashlib
import os
import subprocess
import sys
from typing import Dict, Optional, Tuple

from devchat.utils import get_logger

from .envs import MAMBA_BIN_PATH
from .path import ENV_CACHE_DIR, MAMBA_PY_ENVS, MAMBA_ROOT
from .schema import ExternalPyConf
from .user_setting import USER_SETTINGS

# CONDA_FORGE = [
#     "https://mirrors.tuna.tsinghua.edu.cn/anaconda/cloud/conda-forge/",
#     "conda-forge",
# ]
CONDA_FORGE_TUNA = "https://mirrors.tuna.tsinghua.edu.cn/anaconda/cloud/conda-forge/"
PYPI_TUNA = "https://pypi.tuna.tsinghua.edu.cn/simple"

logger = get_logger(__name__)


def _get_external_envs() -> Dict[str, ExternalPyConf]:
    """
    Get the external python environments info from the user settings.
    """
    external_pythons: Dict[str, ExternalPyConf] = {}
    for conf in USER_SETTINGS.external_workflow_python:
        external_pythons[conf.env_name] = conf

    return external_pythons


EXTERNAL_ENVS = _get_external_envs()


class PyEnvManager:
    mamba_bin = MAMBA_BIN_PATH
    mamba_root = MAMBA_ROOT

    def __init__(self):
        pass

    @staticmethod
    def get_py_version(py: str) -> Optional[str]:
        """
        Get the version of the python executable.
        """
        py_version_cmd = [py, "--version"]
        with subprocess.Popen(py_version_cmd, stdout=subprocess.PIPE, stderr=None) as proc:
            proc.wait()

            if proc.returncode != 0:
                return None

            out = proc.stdout.read().decode("utf-8")
            return out.split()[1]

    @staticmethod
    def get_dep_hash(reqirements_file: str) -> str:
        """
        Get the hash of the requirements file content.

        Used to check if the requirements file has been changed.
        """
        with open(reqirements_file, "r", encoding="utf-8") as f:
            content = f.read()
            return hashlib.md5(content.encode("utf-8")).hexdigest()

    def ensure(
        self, env_name: str, py_version: str, reqirements_file: Optional[str] = None
    ) -> Optional[str]:
        """
        Ensure the python environment exists with the given name and version.
        And install the requirements if provided.

        return the python executable path.
        """
        py = self.get_py(env_name)

        should_remove_old = False

        if py:
            # check the version of the python executable
            current_version = self.get_py_version(py)

            if current_version != py_version:
                should_remove_old = True

            if reqirements_file and self.should_reinstall(env_name, reqirements_file):
                should_remove_old = True

            if not should_remove_old:
                return py

        print("\n```Step\n# Setting up workflow environment\n", flush=True)
        if should_remove_old:
            print(f"- Dependencies of {env_name} have been changed.", flush=True)
            print(f"- Removing the old {env_name}...", flush=True)
            self.remove(env_name)

        # create the environment
        print(f"- Creating {env_name} with {py_version}...", flush=True)
        create_ok, msg = self.create(env_name, py_version)
        if not create_ok:
            print(f"- Failed to create {env_name}", flush=True)
            print("\n```", flush=True)
            print(
                f"\n\nFailed to create {env_name}, the workflow will not run."
                f"\n\nPlease try again later.",
                flush=True,
            )
            logger.warning(f"Failed to create {env_name}: {msg}")
            sys.exit(0)
            # return None

        # install the requirements
        if reqirements_file:
            filename = os.path.basename(reqirements_file)
            print(f"- Installing dependencies from {filename}...", flush=True)
            install_ok, msg = self.install(env_name, reqirements_file)
            if not install_ok:
                print(f"- Failed to install dependencies from {filename}", flush=True)
                print("\n```", flush=True)
                print(
                    "\n\nFailed to install dependencies, the workflow will not run."
                    "\n\nPlease try again later.",
                    flush=True,
                )
                logger.warning(f"Failed to install dependencies: {msg}")
                sys.exit(0)
                # return None

            # save the hash of the requirements file content
            dep_hash = self.get_dep_hash(reqirements_file)
            cache_file = os.path.join(ENV_CACHE_DIR, f"{env_name}")
            with open(cache_file, "w", encoding="utf-8") as f:
                f.write(dep_hash)

        print("\n```", flush=True)
        return self.get_py(env_name)

    def install(self, env_name: str, requirements_file: str) -> Tuple[bool, str]:
        """
        Install requirements into the python environment.

        Args:
            requirements: the absolute path to the requirements file.
        """
        py = self.get_py(env_name)
        if not py:
            return False, "Python executable not found."

        if not os.path.exists(requirements_file):
            return False, "Dependencies file not found."

        cmd = [
            py,
            "-m",
            "pip",
            "install",
            "-r",
            requirements_file,
            "-i",
            PYPI_TUNA,
            "--no-warn-script-location",
        ]
        env = os.environ.copy()
        env.pop("PYTHONPATH")
        with subprocess.Popen(
            cmd, stdout=subprocess.DEVNULL, stderr=subprocess.PIPE, env=env
        ) as proc:
            _, err = proc.communicate()

            if proc.returncode != 0:
                return False, err.decode("utf-8")

            return True, ""

    def should_reinstall(self, env_name: str, requirements_file: str) -> bool:
        """
        Check if the requirements file has been changed.
        """
        cache_file = os.path.join(ENV_CACHE_DIR, f"{env_name}")
        if not os.path.exists(cache_file):
            return True

        dep_hash = self.get_dep_hash(requirements_file)
        with open(cache_file, "r", encoding="utf-8") as f:
            cache_hash = f.read()

        return dep_hash != cache_hash

    def create(self, env_name: str, py_version: str) -> Tuple[bool, str]:
        """
        Create a new python environment using mamba.
        """
        is_exist = os.path.exists(os.path.join(MAMBA_PY_ENVS, env_name))
        if is_exist:
            return True

        # create the environment
        cmd = [
            self.mamba_bin,
            "create",
            "-n",
            env_name,
            "-c",
            CONDA_FORGE_TUNA,
            "-r",
            self.mamba_root,
            f"python={py_version}",
            "-y",
        ]
        with subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE) as proc:
            out, err = proc.communicate()
            msg = f"err: {err.decode()}\n-----\nout: {out.decode()}"

            if proc.returncode != 0:
                return False, msg
            return True, ""

    def remove(self, env_name: str) -> bool:
        """
        Remove the python environment.
        """
        is_exist = os.path.exists(os.path.join(MAMBA_PY_ENVS, env_name))
        if not is_exist:
            return True

        # remove the environment
        cmd = [
            self.mamba_bin,
            "env",
            "remove",
            "-n",
            env_name,
            "-r",
            self.mamba_root,
            "-y",
        ]
        with subprocess.Popen(cmd, stdout=subprocess.DEVNULL, stderr=None) as proc:
            proc.wait()

            if proc.returncode != 0:
                return False

            return True

    def get_py(self, env_name: str) -> Optional[str]:
        """
        Get the python executable path of the given environment.
        """
        env_path = None
        if sys.platform == "win32":
            env_path = os.path.join(MAMBA_PY_ENVS, env_name, "python.exe")
            # env_path = os.path.join(MAMBA_PY_ENVS, env_name, "Scripts", "python.exe")
        else:
            env_path = os.path.join(MAMBA_PY_ENVS, env_name, "bin", "python")

        if env_path and os.path.exists(env_path):
            return env_path

        return None
